@@ -957,11 +957,6 @@ coreconfigitem(
     b'directaccess.revnums',
     default=False,
 )
-coreconfigitem(
-    b'experimental',
-    b'dirstate-tree.in-memory',
-    default=False,
-)
 coreconfigitem(
     b'experimental',
     b'editortmpinhg',
@@ -1264,6 +1259,11 @@ coreconfigitem(
     b'web.api.debugreflect',
     default=False,
 )
+coreconfigitem(
+    b'experimental',
+    b'web.full-garbage-collection-rate',
+    default=1,  # still forcing a full collection on each request
+)
 coreconfigitem(
     b'experimental',
     b'worker.wdir-get-thread-safe',
@@ -1306,9 +1306,10 @@ coreconfigitem(
     # Enable this dirstate format *when creating a new repository*.
     # Which format to use for existing repos is controlled by .hg/requires
     b'format',
-    b'exp-dirstate-v2',
+    b'use-dirstate-v2',
     default=False,
     experimental=True,
+    alias=[(b'format', b'exp-rc-dirstate-v2')],
 )
 coreconfigitem(
     b'format',
@@ -1880,6 +1881,13 @@ coreconfigitem(
     default=b'skip',
     experimental=True,
 )
+# experimental as long as format.use-dirstate-v2 is.
+coreconfigitem(
+    b'storage',
+    b'dirstate-v2.slow-path',
+    default=b"abort",
+    experimental=True,
+)
 coreconfigitem(
     b'storage',
     b'new-repo-backend',
